/*
** Module   : BrowseWidget.java
** Abstract : server side BROWSE widget class
**
** Copyright (c) 2005-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 NVS 20051026   @23481 Created initial version.
** 002 SIY 20060307   @25125 Major update for complete BROWSE support.
** 003 SIY 20060319   @25125 Fixed update.
** 004 SIY 20060328   @25276 Fixed reposition code, added a number of 
**                           methods to set various flags and attributes.
** 004 SIY 20060406   @25400 Added setCentered() methods. 
** 005 SIY 20060427   @25816 Added another signature for 
**                           setRepositionerRow() method.
** 006 ECF 20060501   @25917 Change to getRowCount. PreselectQueryWrapper
**                           is now obsolete; use new P2JQuery.isPreselect
**                           method instead.
** 007 ECF 20060618   @27424 Change to setQuery(). Ensure new query is set
**                           to be a scrolling query.
** 008 GES 20060803   @28366 Some cleanup and refactoring.
** 009 GES 20060804   @28430 Honor new iterate() processing which only
**                           happens at the top of the block body.
** 010 GES 20060808   @28490 Use int instead of long for the backing 
**                           methods that support the server exports.
** 011 ECF 20061019   @30532 Changed reposition(). Removed adjustment code
**                           for query off-end case.
** 012 ECF 20061023   @30610 Made BrowseWidget tolerant of a query which
**                           returns no results. When a query is set into
**                           the widget, its errorIfNull property is set
**                           to false. QueryOffEndException is caught and
**                           ignored in finishSetup() when trying to
**                           retrieve the first record.
** 013 ECF 20061103   @30963 Implement RepositionListener interface. This
**                           is necessary to received notifications of
**                           reposition events from the associated query.
**                           Made additional changes to support changes
**                           in the P2JQuery interface, and to support the
**                           assumption that a new record is fetched when
**                           the associated query is repositioned.
** 014 SIY 20061105   @30983 Finished support for notifications and added
**                           getCurrentRow() method.
** 015 GES 20061110   @31077 Fixed abend in getCurrentRow().
** 016 SIY 20061114   @31360 Fixed some repositioning issues. Removed 
**                           initial query positioning which did break 
**                           browse behavior (initially selected row is
**                           the current row in query).
** 017 ECF 20061129   @31479 Fixed getRowCount(). This was aborting with a
**                           QueryOffEndException in the case of a query
**                           which had no results. This exception is now
**                           caught and handled.
** 018 ECF 20061204   @31578 Use new P2JQuery.setBrowsed() method. Also,
**                           getCurrentRow() reports -1 instead of 0 for
**                           current row if query is off-end, since 0 is a
**                           valid row number.
** 019 SIY 20070316   @32421 Set column ordinal when it is added to browse.
** 020 SIY 20070507   @33425 Changed handling of record set size request 
**                           and other relevant changes.
** 021 SIY 20070515   @33755 Fixed positioning. 
** 022 SIY 20070605   @33963 Fixed restoring of the position after getRows.
** 023 NVS 20070618   @34083 Fixed ignoreRepo state variable rebalancing.
**                           Due to exceptions in finally{} clauses it
**                           could have been left unbalanced at times,
**                           causing legitimate requests to be ignored.
** 024 NVS 20070622   @34206 Fixed regression introduced by #022.
**                           reposition() calls throw QueryOffEndException
**                           when the result set is empty in getRows().
** 025 ECF 20070928   @35274 Fixed setCurrentRow(). Was repositioning to
**                           first row after running off-end at last row.
** 026 ECF 20071011   @35446 Further adjustment to setCurrentRow(). Avoid
**                           multiple repositions in case of selecting
**                           last browse row.
** 027 ECF 20080208   @37029 Fixed fetchSelectedRow(). This method was
**                           improperly using its argument as the row
**                           index, rather than as an index within the
**                           list of selected rows.
** 028 GES 20080320   @37581 Added TYPE attribute.
** 029 ECF 20080509   @38246 Added safety code to getRows(). We must first
**                           check if any rows are available before trying
**                           to fetch them.
** 030 SVL 20080512   @38277 setScrolling() and setErrorIfNull(false) were
**                           removed from setQuery(), reposition(1) was
**                           added to setQuery().
** 031 SVL 20080604   @38536 currentRowImpl() is used instead of
**                           currentRow().
** 032 SIY 20080625   @38923 Minor fixes in javadoc and removed unnecessary
**                           type cast.
** 033 CA  20080925   @39961 Support for dynamic title.
** 034 ECF 20081106   @40318 Improved getRows(). Use a dynamic list
**                           instead of a fixed array, in case the row
**                           count changes mid-query. Break fetch loop on
**                           query off-end condition.
** 035 CA  20081203   @40813 getColumns() must set the ID of the browse 
**                           column, before sending it to the client.
** 036 CA  20081210   @40830 Events triggered via APPLY will force the
**                           buffer to load the currently selected row.
** 037 CA  20081210   @40842 Implemented CLOSE QUERY statement. On query
**                           close, the browse will reset its data.
** 038 CA  20081212   @40886 Fixed a regression in setCurrentRow, caused by
**                           #036: rowIndex must always be incemented (so it
**                           will be 1-based).
** 039 SIY 20090213   @41306 Added 'force' parameter into refreshStartRow
**                           signature. Added forceRefresh() method.
** 040 SIY 20090706   @43055 Removed apply(). Changes in query behavior made
**                           it obsolete.
** 041 CA  20090906   @43831 Fixed FETCH-SELECTED-ROW() - an error message 
**                           must be shown if the received index is unknown.
**                           Fixed SET-REPOSITIONED-ROW() - if the received
**                           row number is unknown, it must return false.
** 042 SVL 20091030   @44295 Offend state is preserved while execution of the 
**                           getRowCount().
** 043 SVL 20091105   @44316 Into getRowCount() set the query cursor final
**                           state to "not between rows" (if the query allows
**                           fetch on reposition).
** 044 SIY 20101122          Minor cleanups.
** 045 SIY 20111101          Updated imports.
** 046 SIY 20111202          Updated imports.
** 047 CA  20121022          Added support for optional query and column list 
**                           clauses.
*/

package com.goldencode.p2j.ui;

import java.util.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.ui.client.*;

/**
 * Server side representation of the Browse widget.
 */
public class BrowseWidget
extends BaseEntity
implements RepositionListener
{
   /** Underlying query.  May be null. */
   private P2JQuery query = null;

   /** The registered storage for the current browse title. */
   private CharacterExpression dynTitle = null;

   /** Convenience reference to config. */
   private BrowseConfig bconfig = null;
   
   /** List of columns in browse. */
   private ArrayList<BrowseColumnWidget> columns =
      new ArrayList<BrowseColumnWidget>();
   
   /** Widget initialization flag. */
   private boolean initDone = false;
   
   /** Flag to ignore browse-driven reposition events */
   private int ignoreRepo = 0;
   
   /**
    * Default constructor.
    */
   public BrowseWidget()
   { 
      super(new BrowseConfig(ComponentConfig.BROWSE, null));
      bconfig = (BrowseConfig) config;
   }
   
   /**
    * Get the HELP writable attribute.
    * 
    * @return   The current value of the HELP attribute.
    */
   public character getHelp()
   {
      String help = bconfig.getHelp();
      return (help == null) ? new character() : new character(help);
   }

   /**
    * Get the MULTIPLE attribute.
    * 
    * @return   The value of the MULTIPLE attribute.
    */
   public logical isMultiple()
   {
      return new logical(bconfig.isMultiple());
   }

   /**
    * Set value of the MULTIPLE attribute.
    * 
    * @param    multiple
    *           New value of the attribute.
    */
   public void setMultiple(boolean multiple)
   {
      bconfig.setMultiple(multiple);
      frame.pushScreenDefinition();
   }
   
   /**
    * Set value of the MULTIPLE attribute.
    * 
    * @param    multiple
    *           New value of the attribute.
    */
   public void setMultiple(logical multiple)
   {
      if (multiple.isUnknown())
         return;
      
      setMultiple(multiple.booleanValue());
   }

   /**
    * Set value of the SCROLLBAR-VERTICAL attribute.
    * 
    * @param    flag
    *           New value of the attribute.
    */
   public void setScrollbarVertical(boolean flag)
   {
      bconfig.setScrollbarVertical(flag);
      frame.pushScreenDefinition();
   }

   /**
    * Set value of the SCROLLBAR-VERTICAL attribute.
    * 
    * @param    flag
    *           New value of the attribute.
    */
   public void setScrollbarVertical(logical flag)
   {
      if (flag.isUnknown())
         return;
      
      setScrollbarVertical(flag.booleanValue());
   }
   
   /**
    * Set OVERLAY attribute.
    * 
    * @param    overlay
    *           New attribute value.
    */
   public void setOverlay(boolean overlay)
   {
      bconfig.setOverlay(overlay);
      frame.pushScreenDefinition();
   }
   
   /**
    * Set OVERLAY attribute.
    * 
    * @param    overlay
    *           New attribute value.
    */
   public void setOverlay(logical overlay)
   {
      if (overlay.isUnknown())
         return;
      
      setOverlay(overlay.booleanValue());
   }

   /**
    * Set value of the ROW-MARKERS attribute.
    * 
    * @param    markers
    *           New value of the attribute.
    */
   public void setRowMarkers(boolean markers)
   {
      bconfig.setRowMarkers(markers);
      frame.pushScreenDefinition();
   }
   
   /**
    * Set value of the ROW-MARKERS attribute.
    * 
    * @param    markers
    *           New value of the attribute.
    */
   public void setRowMarkers(logical markers)
   {
      if (markers.isUnknown())
         return;
      
      setRowMarkers(markers.booleanValue());
   }

   /**
    * Set value of the ROW-MARKERS attribute as opposite to passed value.
    * 
    * @param    nomarkers
    *           New value of the attribute.
    */
   public void setNoRowMarkers(boolean nomarkers)
   {
      setRowMarkers(!nomarkers);
   }
   
   /**
    * Set value of the ROW-MARKERS attribute as opposite to passed value.
    * 
    * @param    nomarkers
    *           New value of the attribute.
    */
   public void setNoRowMarkers(logical nomarkers)
   {
      if (nomarkers.isUnknown())
         return;
      
      setNoRowMarkers(nomarkers.booleanValue());
   }
   
   /**
    * Set SEPARATORS attribute as opposite to passed value.
    * 
    * @param    noseparators
    *           New value for the attribute.
    */
   public void setNoSeparators(boolean noseparators)
   {
      bconfig.setSeparators(!noseparators);
      frame.pushScreenDefinition();
   }

   /**
    * Set SEPARATORS attribute as opposite to passed value.
    * 
    * @param    noseparators
    *           New value for the attribute.
    */
   public void setNoSeparators(logical noseparators)
   {
      if (noseparators.isUnknown())
         return;
      
      setNoSeparators(noseparators.getValue());
   }
   
   /**
    * Set SEPARATORS attribute.
    * 
    * @param    separators
    *           New value for the attribute.
    */
   public void setSeparators(boolean separators)
   {
      bconfig.setSeparators(separators);
      frame.pushScreenDefinition();
   }

   /**
    * Set SEPARATORS attribute.
    * 
    * @param    separators
    *           New value for the attribute.
    */
   public void setSeparators(logical separators)
   {
      if (separators.isUnknown())
         return;
      
      setSeparators(separators.booleanValue());
   }

   /**
    * Set LABELS attribute.
    * 
    * @param    labels
    *           New value for the attribute.
    */
   public void setLabels(boolean labels)
   {
      bconfig.setLabels(labels);
      frame.pushScreenDefinition();
   }

   /**
    * Set LABELS attribute.
    * 
    * @param    labels
    *           New value for the attribute.
    */
   public void setLabels(logical labels)
   {
      if (labels.isUnknown())
         return;
      
      setLabels(labels.booleanValue());
   }

   /**
    * Set LABELS attribute as opposite to passed value.
    * 
    * @param    nolabels
    *           New value for the attribute.
    */
   public void setNoLabels(boolean nolabels)
   {
      setLabels(!nolabels);
   }

   /**
    * Set CENTERED attribute value.
    * 
    * @param    centered
    *           New value for the attribute.
    */
   public void setCentered(boolean centered)
   {
      bconfig.setCentered(centered);
      frame.pushScreenDefinition();
   }

   /**
    * Set CENTERED attribute value.
    * 
    * @param    centered
    *           New value for the attribute.
    */
   public void setCentered(logical centered)
   {
      if (centered.isUnknown())
         return;
      
      setCentered(centered.booleanValue());
   }
   
   /**
    * Set LABELS attribute as opposite to passed value.
    * 
    * @param    labels
    *           New value for the attribute.
    */
   public void setNoLabels(logical labels)
   {
      if (labels.isUnknown())
         return;
      
      setNoLabels(labels.booleanValue());
   }

   /**
    * Set BOX attribute.
    * 
    * @param    box
    *           New value for the attribute.
    */
   public void setBox(boolean box)
   {
      bconfig.setBox(box);
      frame.pushScreenDefinition();
   }

   /**
    * Set BOX attribute.
    * 
    * @param    box
    *           New value for the attribute.
    */
   public void setBox(logical box)
   {
      if (box.isUnknown())
         return;
      
      setBox(box.booleanValue());
   }

   /**
    * Set BOX attribute as opposite to passed value.
    * 
    * @param    nobox
    *           New value for the attribute.
    */
   public void setNoBox(boolean nobox)
   {
      setBox(!nobox);
   }

   /**
    * Set BOX attribute as opposite to passed value.
    * 
    * @param    nobox
    *           New value for the attribute.
    */
   public void setNoBox(logical nobox)
   {
      if (nobox.isUnknown())
         return;
      
      setNoBox(nobox.booleanValue());
   }

   /**
    * Set value of the MULTIPLE attribute as opposite to passed argument.
    * 
    * @param    single
    *           New value of the attribute.
    */
   public void setSingle(logical single)
   {
      if (single.isUnknown())
         return;
      
      setSingle(single.booleanValue());
   }

   /**
    * Set value of the MULTIPLE attribute as opposite to passed argument.
    * 
    * @param    single
    *           New value of the attribute.
    */
   public void setSingle(boolean single)
   {
      setMultiple(!single);
   }

   /**
    * Set browse widget width.
    * 
    * @param    width
    *           New width.
    */
   public void setWidth(int width)
   {
      bconfig.setWidth(width);
      frame.pushScreenDefinition();
   }

   /**
    * Set browse widget width.
    * 
    * @param    width
    *           New width.
    */
   public void setWidth(integer width)
   {
      if (width.isUnknown())
         return;
      
      setWidth(width.intValue());
   }

   /**
    * Sets the dynamic title for this browse widget.
    *
    * @param    title
    *           The new value for the dynamic title.
    */
   public void setDynamicTitle(BaseDataType title)
   {
      setDynamicTitle(new AccessorWrapper(title));
   }
   
   /**
    * Sets the dynamic title for this browse widget.
    *
    * @param    title
    *           The new value for the dynamic title.
    */
   public void setDynamicTitle(Accessor title) 
   {
      class TitleExpr
      extends CharacterExpression
      {
         private final Accessor titleExpr;
         
         /**
          * Construct an instance for given {@link Accessor} instance.
          * 
          * @param    titleExpr
          *           Dynamic title variable.
          */
         public TitleExpr(Accessor titleExpr)
         {
            this.titleExpr = titleExpr;
         }
         
         public character execute()
         {
            return (character) titleExpr.get();
         }
      }
      setDynamicTitle(new TitleExpr(title));
   }

   /**
    * Sets the dynamic title for this browse widget.
    *
    * @param    titleExp
    *           The new expression for the dynamic title.
    */
   public void setDynamicTitle(CharacterExpression titleExp) 
   {
      this.dynTitle = titleExp;

      // the current value computed from the dynamic title is used until 
      // the buffer is sent to the client and the title is updated
      character title = dynTitle.execute();
      String    txt   = (title.isUnknown() ? "" : title.toStringMessage());
      base.setTitle(txt);
   }

   /**
    * Sets the TITLE writable attribute.
    *
    * @param    title
    *           The new value for the TITLE attribute.
    */
   public void setTitle(String title)
   {
      super.setTitle(title);
      
      // if the browse TITLE attribute is set, then the dynamic title
      // must be canceled
      this.dynTitle = null;
   }

   /**
    * Get value of NUM-COLUMNS attribute.
    * 
    * @return   Number of columns in the browse.
    */
   public integer getNumColumns()
   {
      return new integer(columns.size());
   }

   /**
    * Get specified browse column.
    * 
    * @param    index
    *           1-based column index.
    *           
    * @return   reference to column widget.
    */
   public BrowseColumnWidget getBrowseColumn(double index)
   {
      BrowseColumnWidget res = columns.get((int) index - 1);
      
      if (res == null)
      {
         ErrorManager.recordOrShowError(4065, 
             "**The GET-BROWSE-COLUMN attribute on the BROWSE " + 
             frame.getName(this) + " has invalid arguments.");
      }
      
      return res;
   }

   /**
    * Get specified browse column.
    * 
    * @param    index
    *           1-based column index.
    *           
    * @return   reference to column widget.
    */
   public BrowseColumnWidget getBrowseColumn(NumberType index)
   {
      double ndx = (index.isUnknown()) ? 0 : index.doubleValue();
      
      return getBrowseColumn(ndx);
   }
   
   /**
    * Get the NUM-SELECTED-ROWS attribute.
    * 
    * @return   The current value of NUM-SELECTED-ROWS attribute.
    */
   public integer getNumSelectedRows()
   {
      return new integer(LogicalTerminal.getNumSelectedRows(this));
   }

   /**
    * Get the QUERY writable attribute.
    * 
    * @return   Current value of the QUERY attribute.
    */
   public P2JQuery getQuery()
   {
      return query;
   }

   /**
    * Set the query which will back this browse.  Ensure it is scrollable and
    * does not raise an error condition if it goes off-end.
    * 
    * @param    query
    *           New value for the attribute.
    */
   public void setQuery(P2JQuery query)
   {
      this.query = query;
      query.setBrowsed(true);
      query.setFetchOnReposition(true);
      query.addRepositionListener(this);
      query.reposition(1);
   }

   /**
    * Implements the SELECT-ALL() widget method.
    * 
    * @return   Always <code>true</code>.
    */
   public logical selectAll()
   {
      LogicalTerminal.selectAll(this);
      return new logical(true);
   }

   /**
    * Implements the DESELECT-ROWS() widget method.
    * 
    * @return   Always <code>true</code>.
    */
   public logical deselectRows()
   {
      LogicalTerminal.deselectRows(this);
      return new logical(true);
   }

   /**
    * Implements the FETCH-SELECTED-ROW() widget method.
    * 
    * @param   n
    *          1-based index of the selected row to fetch;  note that this is
    *          the index into the list of selected rows within the browse,
    *          NOT the index of the row itself.
    *
    * @return  <code>true</code> if method succeeds.
    */
   public logical fetchSelectedRow(integer n)
   {
      if (query == null)
      {
         final String msg = "Browse widget must contain at least one record" + 
                            " before FETCH-SELECTED-ROW method is used";
         ErrorManager.displayError(2108, msg, false);

         return new logical(false);
      }

      if (n.isUnknown())
      {
         final String msg = "Invalid row index specified for browse method " + 
                            "FETCH-SELECTED-ROW";
         ErrorManager.displayError(385, msg, false);

         return new logical(false);
      }
      
      int selListIndex = n.intValue() - 1;
      if (selListIndex < 0)
         return new logical(false);
      
      // Get the 0-based row index of the (n - 1)th selected row.
      int uiRow = LogicalTerminal.getSelectedRowIndex(this, selListIndex);
      if (uiRow < 0)
         return new logical(false);
      
      // Reposition the query to this row and update the underlying buffers.
      reposition(new integer(uiRow + 1), true);
      
      if (query.isOffEnd().booleanValue())
         return new logical(false);
      
      return new logical(true);
   }

   /**
    * Implements the REFRESH() widget method.
    * 
    * @return   Always <code>true</code>.
    */
   public logical refresh()
   {
      LogicalTerminal.refresh(this);
      return new logical(true);
   }

   /**
    * Implements the SELECT-FOCUSED-ROW() widget method.
    * 
    * @return   Always <code>true</code>.
    */
   public logical selectFocusedRow()
   {
      LogicalTerminal.selectFocusedRow(this);
      return new logical(true);
   }

   /**
    * Implements the SET-REPOSITIONED-ROW() widget method.
    * 
    * @param    row
    *           1-based row number where the new record is displayed
    * @param    cond
    *           if <code>true</code>, tries to reposition to the browse
    *           viewport existing row first
    * 
    * @return   Always <code>true</code>.
    */
   public logical setRepositionedRow(integer row, boolean cond)
   {
      if (row.isUnknown())
      {
         return new logical(false);
      }

      LogicalTerminal.setRepositionedRow(this, row.intValue() - 1, cond);
      return new logical(true);
   }

   /**
    * Implements the SET-REPOSITIONED-ROW() widget method.
    * 
    * @param    row
    *           1-based row number where the new record is displayed
    * @param    string
    *           If this parameter is "conditional", then method tries to
    *           reposition to the browse viewport existing row first
    * 
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   public logical setRepositionedRow(integer row, String string)
   {
      boolean conditional = "conditional".startsWith(string.toLowerCase()); 

      return setRepositionedRow(row, conditional);
   }
   
   /**
    * Blocks attempts to set the DCOLOR attribute as it is read-only.
    * 
    * @param   d
    *          Flag is not used.
    */
   public void setDcolor(integer d)
   {
      throw new RuntimeException("DCOLOR attribute is read only");
   }

   /**
    * Blocks attempts to get the PFCOLOR attribute as it is undefined.
    * 
    * @return  nothing 
    */
   public integer getPfColor()
   {
      throw new RuntimeException("PFCOLOR attribute is undefined");
   }

   /**
    * Blocks attempts to set the PFCOLOR attribute as it is undefined.
    * 
    * @param   p
    *          Flag is not used.
    */
   public void setPfColor(integer p)
   {
      throw new RuntimeException("PFCOLOR attribute is undefined");
   }
   
   // ------------------------------------------------------------------
   // Client side BrowseModel request handlers
   // ------------------------------------------------------------------

   /**
    * Return information about columns.
    * 
    * @return   Meta information about columns.
    */
   public BrowseColumnConfig[] getColumns()
   {
      BrowseColumnConfig[] res = new BrowseColumnConfig[columns.size()];
      int i = 0;
      
      for (BrowseColumnWidget column : columns)
      {
         BrowseColumnConfig cfg = 
            new BrowseColumnConfig(ComponentConfig.BROWSE_COLUMN, null);
         
         cfg.applyConfig(column.getConfig());
         cfg.setID(column.getConfig().getID());
         
         res[i++] = cfg;
      }
      
      return res;
   }

   /**
    * Get number of available rows starting from the given one.
    * 
    * @param   startRow
    *          Initial row to start count rows from. 
    * 
    * @return  Number of available rows or 0 if no query specified.
    */
   public int getRowCount(int startRow)
   {
      if (query == null)
      {
         return 0;
      }

      int res = 0;
      
      integer rowId = query.currentRowImpl();
      boolean isFoR = query.isFetchOnReposition();
      OffEnd offEnd = query.getOffEnd();

      try
      {
         ignoreRepo++;
         
         query.setFetchOnReposition(false);
         
         if (query.isPreselect())
         {
            query.first();
            res = query.size().intValue();
         }
         else
         {
            int opos = (rowId.isUnknown() ? -1 : rowId.intValue() - 1);
            if (opos >= 0)
            {
               query.forward(Browse.CACHE_SIZE * 5 + (startRow - opos));
               res = getCurrentRow();   
            }
         }
      }
      catch (QueryOffEndException exc)
      {
         // Thrown if query.first() cannot move to a first result because the
         // result set is empty.
      }
      finally
      {
         ignoreRepo--;
         query.setFetchOnReposition(isFoR);

         if (!rowId.isUnknown())
         {
            reposition(rowId, true);
         }
         else if (offEnd == OffEnd.BACK)
         {
            query.last();
         }
         else if (res > 0)
         {
            query.first();
         }
      }
      
      return res;
   }
   
   /**
    * Retrieves several subsequent rows from the database.
    * 
    * @param    startRowIndex
    *           Index of the first row to retrieve.
    * @param    count
    *           Maximum count of rows to return.
    * 
    * @return   Array of rows. Each row is array of cell values.
    *           If no query specified, return an array of size 0.
    */
   public BaseDataType[][] getRows(int startRowIndex, int count)
   {
      if (query == null)
      {
         return new BaseDataType[0][0];
      }

      int rowCount = getRowCount(startRowIndex);
      count = Math.min(rowCount, count);

      ArrayList<BaseDataType[]> data = new ArrayList<BaseDataType[]>(count);
      
      if (count == 0)
      {
         return (new BaseDataType[0][columns.size()]);
      }

      integer rowId = query.currentRowImpl();
      boolean fOr = query.isFetchOnReposition();
      
      try
      {
         ignoreRepo++;

         reposition(new integer(startRowIndex + 1), false);
         query.setFetchOnReposition(true);
         
         for (int i = 0; i < count; i++)
         {
            query.next();
            
            if (query._isOffEnd())
            {
               break;
            }
            
            BaseDataType[] cols = new BaseDataType[columns.size()];
            data.add(cols);
            
            Iterator<BrowseColumnWidget> iter = columns.iterator();
            for (int j = 0; iter.hasNext(); j++)
            {
               BrowseColumnWidget column = iter.next();
               BaseDataType val = column.getValue();
               
               cols[j] = val;
            }
         }
      }
      catch (QueryOffEndException exc)
      {
         // potentially, may be thrown in reposition() because the result set
         // is empty.
      }
      finally
      {
         ignoreRepo--;

         try
         {
            // Now reposition to where we were originally.
            if (!rowId.isUnknown())
               reposition(rowId, true);
            else
               query.first();
         }
         catch (QueryOffEndException exc)
         {
            // thrown in reposition() if the result set is empty.
         }
         finally
         {
            query.setFetchOnReposition(fOr);
         }
      }
      
      BaseDataType[][] result = data.toArray(new BaseDataType[data.size()][]);

      return result;
   }
   
   /**
    * Updates selected cells in the given row or nothing to do if no query
    * specified.
    * 
    * @param    rowIndex
    *           Index of the row to update.
    * @param    cellIndices
    *           Cells for update indexes.
    * @param    cellValues
    *           Cells for update new values.
    * 
    * @throws   DataValidationException
    *           New values did not pass database validation rules.
    */
   public void updateRow(int            rowIndex,
                         int[]          cellIndices,
                         BaseDataType[] cellValues)
   throws DataValidationException
   {
      if (query == null)
      {
         return;
      }

      integer rowId = query.currentRowImpl();
      String msg = null;
      
      reposition(new integer(rowIndex + 1), true);

      BrowseColumnWidget[] c = new BrowseColumnWidget[columns.size()];
      
      c = columns.toArray(c);
      
      TransactionManager.pushScope("update",
                                   TransactionManager.TRANSACTION,
                                   false,
                                   false,
                                   false,
                                   false);
      try
      {
         update:
         {
            TransactionManager.blockSetup();
            
            do
            {
               try
               {
                  for (int i = 0; i < cellIndices.length; i++)
                  {
                     c[cellIndices[i]].setValue(cellValues[i]);
                  }
               }
               catch (ErrorConditionException err_except)
               {
                  TransactionManager.honorError(err_except);
                  TransactionManager.rollback("update");
                  TransactionManager.triggerRetry(null);
                  msg = err_except.getMessage();
               }
               catch (RetryUnwindException ru_except)
               {
                  TransactionManager.honorRetry(ru_except);
               }
            }
            while (TransactionManager.needsRetry());
            
            if (TransactionManager.isBreakPending())
            {
               break update;
            }
         }
      }
      finally
      {
         TransactionManager.popScope();
      }      
      
      reposition(rowId, true);
      
      if (msg != null)
         throw new DataValidationException(msg);
   }

   /**
    * Returns <code>true</code> if specified row is new and is not present
    * in the data base.
    * 
    * @param    rowIndex
    *           Index of the row to return information about.
    *          
    * @return   <code>true</code> if row is new and <code>false</code>
    *           otherwise.
    */
   public boolean isNewRow(int rowIndex)
   {
      if (query == null)
      {
         return false;
      }

      integer save = query.currentRowImpl();
      
      reposition(new integer(rowIndex + 1), false);

      boolean res = query.isOffEnd().booleanValue(); 
      
      reposition(save, false);

      return res;
   }

   /**
    * Set new current row in the browse. This method is called as notification
    * from the client that current row is changed.
    * 
    * @param   rowIndex
    *          Index of the current row. This index is 0-based and generally
    *          is non-negative, except for the special case of selecting the
    *          very last row in the browse.
    * @param   force
    *          Force positioning to the specified row, even if it is selected.          
    *          
    * @return  New current row or 0 if no query specified.
    */
   public int setCurrentRow(int rowIndex, boolean force)
   {
      if (query == null)
      {
         return 0;
      }

      if (!force && rowIndex == getCurrentRow())
      {
         return rowIndex;
      }
      
      ignoreRepo++;
      boolean fOr = query.isFetchOnReposition();
      
      try
      {
         if (rowIndex < 0)
         {
            // Special case:  move to last record in browse.
            query.setFetchOnReposition(true);
            query.last();
         }
         else
         {
            query.setFetchOnReposition(true);
            reposition(new integer(rowIndex + 1), true);
            logical offEnd = query.isOffEnd();
            if (!offEnd.isUnknown() && offEnd.booleanValue())
            {
               // If we've gone too far, reset ourselves to the last record.
               query.last();
            }
         }
      }
      catch (QueryOffEndException exc)
      {
         // Thrown if result set is empty.
      }
      finally
      {
         ignoreRepo--;
         query.setFetchOnReposition(fOr);
      }
      
      return getCurrentRow();
   }

   /**
    * Get current row in the query.
    *
    * @return   The current row index (0-based) or -1 if the query is off-end.
    */
   public int getCurrentRow()
   {
      if (query == null)
      {
         return -1;
      }

      integer row = query.currentRowImpl();
      return (row.isUnknown() ? -1 : row.intValue() - 1);
   }
   
   //---------------------------------------------------------------
   
   /**
    * Set new value for the DOWN attribute.
    * 
    * @param    down
    *           New value for the attribute.
    */
   public void setDown(int down)
   {
      bconfig.setDown(down);
      frame.pushScreenDefinition();
   }

   /**
    * Add new field to the list of browse fields.
    * 
    * @param    column
    *           Field to add.
    */
   public void addColumn(BrowseColumnWidget column)
   {
      ((BrowseColumnConfig) column.getConfig()).setOrdinal(columns.size());
      columns.add(column);
      column.setBrowseId(getId());
   } 

   /**
    * Change editable state for all columns.
    * 
    * @param    editable
    *           New value for the attribute.
    */
   public void setEditableAll(boolean editable)
   {
      for (BrowseColumnWidget column : columns)
      {
         column.setEditable(editable);
      }
   }

   /**
    * Register query and list of fields with the browse.
    * 
    * @param    query
    *           Reference to query. May be null.
    * 
    * @param    list
    *           List of fields. May be null.
    */
   public void registerQuery(P2JQuery query, FrameElement[] list)
   {
      if (query != null)
      {
         setQuery(query);
      }

      if (list != null)
      {
         for (int i = 0; i < list.length; i++)
         {
            // provide link
            ((BrowseColumnWidget) list[i].getWidget()).setElement(list[i]);
         }
      }
   }

   /**
    * Finish initialization of the browse widget. 
    */
   public void finishSetup()
   {
      if (initDone)
         return;
    
      if (query == null)
      {
         initDone = true;
         return;
      }
      
      // finish initialization of the widget
      try
      {
         if (query.currentRowImpl().isUnknown())
            query.first();
      }
      catch (QueryOffEndException exc)
      {
         // OK to ignore;  indicates query has no results and browse will be
         // empty initially.
      }
      bconfig.setQuery(getId());
      frame.pushScreenDefinition();
      
      initDone = true;
   }

   /**
    * Invoked when the query with which this browse widget is associated is
    * explicitly repositioned.
    */
   public void queryRepositioned()
   {
      if (ignoreRepo > 0)
      {
         return;
      }

      LogicalTerminal.refreshStartRow(this, false);
   }

   /**
    * Force refreshing of the browse at client side.
    */
   public void forceRefresh()
   {
      if (ignoreRepo > 0)
      {
         return;
      }

      LogicalTerminal.refreshStartRow(this, true);
   }
   
   /**
    * Invoked when the query with which this browse widget is associated is
    * explicitly closed.
    */
   public void queryClosed()
   {
      LogicalTerminal.queryClosed(this);
   }

   /**
    * Obtains the type of widget description.
    *
    * @return   Always "BROWSE".
    */
   public character getType()
   {
      return new character("BROWSE");
   }

   /**
    * Compute the expression kept in the {@link #dynTitle} field and set it
    * as the current browse title. This operation will cancel the old
    * {@link #dynTitle} by setting it to null.
    */
   void updateDynamicTitle()
   {
      if (dynTitle != null)
      {
         character title = dynTitle.execute();
         String    txt   = (title.isUnknown() ? "" : title.toStringMessage());

         setTitle(txt);
      }
   }
   
   /**
    * Common reposition code.
    * 
    * @param   row
    *          1-based row to set.
    * @param   fetch
    *          <code>true</code> to auto-fetch the record after reposition,
    *          <code>false</code> to move the cursor only.
    */
   private void reposition(integer row, boolean fetch)
   {
      if (query == null)
      {
         return;
      }

      boolean fOr = query.isFetchOnReposition(); 
      
      try
      {
         ignoreRepo++;
         
         if (!fetch)
            query.setFetchOnReposition(false);
         
         query.reposition(row);
      }
      finally
      {
         ignoreRepo--;
         
         if (!fetch)
            query.setFetchOnReposition(fOr);
      }
   }
}
