/*
** Module   : BrowseImpl.java
** Abstract :
**
** Copyright (c) 2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20111124          Extracted CHUI-specific part from Browse.
** 002 CA  20121022          Added support for optional query and column-list
**                           clauses.
*/

package com.goldencode.p2j.ui.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * CHUI-specific portion of implementation of Progress BROWSE widget.
 */
@SuppressWarnings("unchecked")
public class BrowseImpl
extends Browse
{
   /** The renderer for columns labels area. */
   private LabelColumnRenderer rndrLabels = null;
   
   /** The renderer for a data area with values. */
   private DataColumnRenderer rndrData = new DataColumnRenderer();
   
   /** The renderer for data lines which are outside data model. */
   private SpaceColumnRenderer rndrSpace = new SpaceColumnRenderer();
   
   /**
    * Constructs the CHUI-specific version of Browse widget.
    * 
    * @param   cfg
    *          Configuration data.
    */
   public BrowseImpl(ComponentConfig cfg)
   {
      super(cfg);
   }

   /**
    * Sets position of terminal's cursor. Redirects this positioning to the
    * in-place editor if it is present. Otherwise does nothing because BROWSE
    * itself has no purpose for the hardware cursor.
    */
   @Override
   public void drawCaret()
   {
      FillIn editor = getActiveEditor();
      
      if (editor !=null)
         editor.drawCaret();
   }

   /**
    * Draws the widget: border, header, selection marks and line, internal
    * FillIn etc.
    */
   @Override
   public void draw()
   {
      // if the model was not initialized yet (i.e. no query was specified),
      // initialize it now.
      if (browseModel == null)
      {
         setQuery(this.config.getID());
      }

      if (config.isRefreshable() == false || browseModel == null)
         return;
      
      if (postponedRefresh)
         refreshStartRow(false);

      Color clr = UiUtils.getDcolor(this);
      colorCommon = clr;
      colorHSep = clr.withUnderline();
      colorCurrent = clr.withReverse();

      // save the default colors, to be restored after drawing
      BrowseColumnConfig[] columns = browseModel.getColumns();
      Color[] oldColors = new Color[columns.length];
      ThinClient client = ThinClient.getInstance();
      for (int i = 0; i < oldColors.length; i++)
      {
         BrowseColumn bcol = (BrowseColumn) client.getWidget(columns[i].getID());
         oldColors[i] = bcol.config().getDcolor();
      }

      // if frame is not alive, do not generate events and clear browse 
      // content
      Widget frame = UiUtils.locateFrame(this);
      
      if (frame != null && !frame.isVisible())
         return;
      
      if (frame instanceof Frame && !((Frame) frame).isAlive())
      {
         resetRowDisplayState();
         this.dcRowCount = 0;
         this.dcData = null;
         this.currRow = -1;   
         this.scrollRow = 0;
         repaint();
      }
      else
      {
         // else, trigger events, before drawing
         triggerRowDisplayEvents(oldColors);
      }

      // save the row from where drawing was started last time
      oldDisplayRow = scrollRow;

      Point origin = screenLocation();
      int x = origin.x;
      int y = origin.y;

      OutputManager screen = screen();
      
      if (config.isBox())
      {
         screen.boxToScreen(origin, size(), clr);
         
         int width = width();

         Color tcl = config.getTitleDcolor();

         if (config.getTitle() != null && config.getTitle().length() > 0)
         {
            String titleText = (config.getTitle().length() > width) ?
                                     config.getTitle().substring(0, width) : 
                                     config.getTitle();
                                     
            int titleOffset = (width - titleText.length()) / 2;
            screen.at(x + titleOffset, y).append(titleText, tcl);
         }
                           
         ++x;
         ++y;
      }

      if (drawLabels)
      {
         if (rcLabels == null)
         {
            if (rndrLabels == null)
               rndrLabels = new LabelColumnRenderer();
            else
               rndrLabels.clear();

            renderAbstractLine(rndrLabels, -1);
            rndrLabels.printResult();
         }

         for (int i = 0; i < ilLabelsHeight; ++i)
         {
            screen.at(x, y).append(rcLabels[i], colorCommon);
            ++y;
         }

         screen.at(x, y).append("  ", colorCommon).hLine(separatorLen , colorCommon);

         if (ilRightSpaceWidth > 0)
         {
            // this cycle usually works only 2 times so it is more effective
            // than to generate special string
            for (int i = 0; i < ilRightSpaceWidth; ++i)
               screen.append(' ', colorCommon);
         }
         
         ++y;
      }

      hadDraw = true;
      
      int rowsLeft = dcRowCount - scrollRow;
      int rowIndex = scrollRow;

      FillIn editor = getActiveEditor();
      boolean readOnly = (editor == null || editor.config().isReadOnly());
      boolean focused = (hasFocus() && isEnabled()) || pseudoFocus;
      
      int rowsOnPageData = Math.min(ilPageRowsCount, rowsLeft);
      config.setNumIterations(rowsOnPageData);
      
      for (; rowsOnPageData > 0; --rowsOnPageData)
      {
         rndrData.newRow(readOnly, focused, rowIndex);
         screen.at(x, y);
         renderAbstractLine(rndrData, rowIndex - scrollRow);
         ++y;
         ++rowIndex;
      }
      
      for (int i = ilPageRowsCount - rowsLeft; i > 0; --i)
      {
         screen.at(x, y);
         rndrSpace.newRow(readOnly, focused, rowIndex);
         renderAbstractLine(rndrSpace, rowIndex - scrollRow);
         ++y;
         ++rowIndex;
      }
      
      if (!focused || editor == null)
      {
         screen.at(0, 0);
      }
      else
      {
         editor.draw();
      }

      // restore the colors to their default values
      for (int k = 0; k < oldColors.length; k++)
      {
         BrowseColumn bcol = (BrowseColumn) client.getWidget(columns[k].getID());
         
         bcol.config().setDcolor(oldColors[k]);
      }
   }


   /**
    * Renders a line from horizontally scrollable area.
    * 
    * @param renderer
    *        Implementation of various rendering's parts.
    * @param rowIdx
    *        The screen index of the row currently being displayed
    */
   private void renderAbstractLine(ColumnRenderer renderer, int rowIdx)
   {
      renderer.addMarks();
      BrowseColumnConfig[] cols = browseModel.getColumns();
      // columns may have not been specified, so exit early if so
      if (cols == null || cols.length == 0)
         return;

      // reset columns
      for (int i = 0; i < cols.length; i++)
         cols[i].setColumn(-1);
      
      if (scrollColumn == 0)
      {
         // no scrolling, the first column is fully visible
         int width = ilPageWidth;
         
         BrowseColumnConfig col = cols[0];
         int colWidth = Math.min(col.getWidthChars(), width);

         if (rowIdx >= 0            &&
             rowIdx < colors.length &&
             colors[rowIdx][0] != null)
         {
            col.setColumnDcolor(colors[rowIdx][0]);
         }
         renderer.addData(col, 0, colWidth, 0);
         width -= colWidth;
         int i;
         
         for (i = 1; i < cols.length; ++i)
         {
            if (width > 0)
            {
               renderer.addSeparator();
               --width;
            }

            if (width <= 0)
            {
               break;
            }
            
            col = cols[i];
            colWidth = col.getWidthChars();
            int scroll = (colWidth > width) ? width - colWidth : 0;

            if (rowIdx >= 0            &&
                rowIdx < colors.length &&
                colors[rowIdx][i] != null)
            {
               col.setColumnDcolor(colors[rowIdx][i]);
            }
            renderer.addData(col, i, colWidth, scroll);
            width -= colWidth;
         }
         
         config.setNumVisibleColumns(i);
      }
      else
      { 
         // scrolling active, scroll column is on the right of view area
         BrowseColumnConfig scol = cols[scrollColumn];
         int nCols = 1;
         int scolWidth = scol.getWidthChars();
         
         if (ilPageWidth <= scolWidth)
         {
            // column is at least as wide as visible area
            scolWidth = ilPageWidth;
         }
         else
         { 
            // view area contains also some columns before scroll column
            int fcolWidth = ilPageWidth - scolWidth - 1;
            int fcolIndex = scrollColumn - 1;
            
            while (true)
            {
               int w = fcolWidth - cols[fcolIndex].getWidthChars() - 1;
               if (w < 0)
               {
                  break;
               }

               fcolWidth = w;
               --fcolIndex;
            }

            BrowseColumnConfig col;
            
            if (fcolWidth > 0) // it may equal 0 also
            {
               col = cols[fcolIndex];
               int fullWidth = col.getWidthChars();

               renderer.addData(col, fcolIndex,
                                fullWidth, fullWidth - fcolWidth);
               nCols++;
            }
            renderer.addSeparator();
            
            for (int i = fcolIndex + 1; i < scrollColumn; ++i)
            {
               col = cols[i];
               renderer.addData(col, i, col.getWidthChars(), 0);
               renderer.addSeparator();
               nCols++;
            }
         }

         renderer.addData(scol, scrollColumn, scolWidth, 0);
         config.setNumVisibleColumns(nCols);
      }
      
      if (ilRightSpaceWidth > 0)
      {
         renderer.addRightSpaces(ilRightSpaceWidth);
      }
   }

   /**
    * Outputs several spaces to the terminal.
    *
    * @param   width
    *          Number of spaces.
    * @param   color
    *          Drawing color.
    */
   private void printSpaces(int width, Color color)
   {
      for (int i = width; i > 0; --i)
      {
         screen().append(' ', color);
      }
   }
   
   /**
    * Interface for renderers of horizontally scrollable rows. The place for
    * saving the result is up to implementation. The result may consist from
    * several character rows.
    */
   private interface ColumnRenderer
   {
      /**
       * Renders data from an one column.
       * 
       * @param   col
       *          Column to render.
       * @param   colIndex
       *          Index of the column.
       * @param   colWidth
       *          The full width for formatting.
       * @param   scroll
       *          Scrolling correction. Means the number of characters which
       *          should be erased from the left (if positive) or from the
       *          right (if negative). This scroll differs from the scroll in
       *          <code>DisplayFormat.formatScreenValue()</code> because it
       *          means the scroll of the whole browse column not the scroll
       *          of screen value inside the column.
       */
      public void addData(BrowseColumnConfig col, int colIndex, int colWidth, 
                          int scroll);

      /**
       * Renders an editable line and a selected line marks.
       */
      public void addMarks();

      /**
       * Renders a separator character.
       */
      public void addSeparator();

      /**
       * Renders several spaces to make right space column.
       * 
       * @param   width
       *          The number of spaces.
       */
      public void addRightSpaces(int width);
   }

   /**
    * Renders the labels area. Places the internal result into the internal
    * buffers. It is expected that <code>clear()</code> method is called
    * before rendering and <code>printResult()</code> is called to finish
    * it.
    */
   private class LabelColumnRenderer
   implements ColumnRenderer
   {
      /** Temporary buffers for render result. */
      private StringBuffer[] bufs;

      /**
       * Constructor. Initializes from the current labels. Object should be
       * recreated if labels are changed.
       */
      public LabelColumnRenderer()
      {
         bufs = new StringBuffer[ilLabelsHeight];
         for (int i = 0; i < bufs.length; ++i)
         {
            bufs[i] = new StringBuffer(ilLabelsHeight);
         }
      }

      /**
       * Clears an old content.
       */
      public void clear()
      {
         for (int i = 0; i < bufs.length; ++i)
         {
            bufs[i].setLength(0);
         }
      }

      /**
       * Copies buffers content into <code>rcLabels</code> member.
       */
      public void printResult()
      {
         rcLabels = new String[bufs.length];
         for (int i = 0; i < bufs.length; ++i)
         {
            rcLabels[i] = bufs[i].toString();
         }
      }

      /**
       * @param   colIndex
       *          Is not used.
       * @see  ColumnRenderer#addData(com.goldencode.p2j.ui.BrowseColumnConfig, int, int, int)
       */
      @Override
      public void addData(BrowseColumnConfig col, int colIndex, int colWidth, 
                          int scroll)
      {
         String[] label = col.getLabelVar();
         
         for (int i = 0; i < label.length; ++i)
         {
            String str = renderCell(label[i], col, colWidth, scroll);
            bufs[i].append(str);
         }

         for (int i = label.length; i < bufs.length; ++i)
         {
            StringBuffer buf = bufs[i];
            for (int j = colWidth - Math.abs(scroll); j > 0; --j)
            {
               buf.append(' ');
            }
         }
      }

      /**
       * @see ColumnRenderer#addMarks()
       */
      @Override
      public void addMarks()
      {
         addRightSpaces(2);
      }

      /**
       * @see ColumnRenderer#addSeparator()
       */
      @Override
      public void addSeparator()
      {
         addRightSpaces(1);
      }

      /**
       * @see ColumnRenderer#addRightSpaces(int)
       */
      @Override
      public void addRightSpaces(int width)
      {
         for (int i = 0; i < bufs.length; ++i)
         {
            for (int j = 0; j < width; ++j)
            {
               bufs[i].append(' ');
            }
         }
      }
   }

   /**
    * Renders a data line which is outside data model. Result is immediately
    * placed to the terminal. No row init necessary.
    */
   private class SpaceColumnRenderer
   implements ColumnRenderer
   {
      /** Indicates read-only mode. */
      protected boolean dcrReadOnly;
      /** Is widget focused. */
      protected boolean dcrFocused;
      /** The index of currently rendered row. */
      protected int dcrRowIndex;

      /** Dummy constructor. */
      public SpaceColumnRenderer(){}

      /**
       * @param   col
       *          Is not used.
       * @param   colIndex
       *          Is not used.
       *          
       * @see  ColumnRenderer#addData(com.goldencode.p2j.ui.BrowseColumnConfig,
       *       int, int, int)
       */
      @Override
      public void addData(BrowseColumnConfig col,
                          int colIndex,
                          int colWidth,
                          int scroll)
      {
         printSpaces(colWidth - Math.abs(scroll), getColor(colorCommon));
      }

      /**
       * Calculate row color.
       *
       * @param   color
       *          Base color.
       * @return  Calculated color.
       */
      public Color getColor(Color color)
      {
         return (dcrReadOnly && dcrRowIndex == Math.max(currRow, 0) && dcrFocused)
                    ? color.withReverse() : (config.isSeparators() ? color.withUnderline() :
                                                       color);
      }

      /**
       * @see  ColumnRenderer#addMarks()
       */
      @Override
      public void addMarks()
      {
         char editMark;
         Color editMarkAttr = colorCommon;
         if (editPossible && config.isRowMarkers())
         {
            editMark = '*';
            
            if (dcrReadOnly && 
                dcrRowIndex == Math.max(currRow, 0) && 
                isEnabled())
            {
               editMarkAttr = colorCurrent;
            }
         }
         else
         {
            editMark = ' ';
         }

         screen().append(editMark, editMarkAttr);
         screen().append(isRowSelected(dcrRowIndex) ? '>' : ' ', colorCommon);
      }

      /**
       * @see  ColumnRenderer#addSeparator()
       */
      @Override
      public void addSeparator()
      {
         Color clr = config.isSeparators() && separatorQuirk ? colorHSep :
                                                               colorCommon;
         char chr = config.isSeparators() ? OutputManager.UACS_VLINE : ' ';

         screen().append(chr, clr);
      }

      /**
       * @see  ColumnRenderer#addRightSpaces(int)
       */
      @Override
      public void addRightSpaces(int width)
      {
         printSpaces(width, colorCommon);
      }

      /**
       * Begins rendering of a new row.
       * 
       * @param   readOnly
       *          The widget in read-only mode.
       * @param   focused
       *          The widget is focused.
       * @param   rowIndex
       *          Index of the row in the browse model.
       */
      public void newRow(boolean readOnly, boolean focused, int rowIndex)
      {
         dcrFocused = focused;
         dcrReadOnly = readOnly;
         dcrRowIndex = rowIndex;
      }
   }

   /**
    * Renders a data line immediately to the terminal. <code>newRow()</code>
    * method should be called before rendering of each row.
    */
   private class DataColumnRenderer
   extends SpaceColumnRenderer
   {
      /** Data to render. */
      protected BaseDataType[] dcrRow;
      
      /**
       * Reference to deferred updates of current row. Always equals
       * <code>currRowUpdates</code> or <code>null</code>.
       */
      protected BaseDataType[] dcrRowUpdates;
      
      /** Dummy constructor. */
      public DataColumnRenderer(){}

      /**
       * @see  SpaceColumnRenderer#newRow(boolean, boolean, int)
       */
      @Override
      public void newRow(boolean readOnly, boolean focused, int rowIndex)
      {
         super.newRow(readOnly, focused, rowIndex);
         
         dcrRow = dcGetRow(rowIndex);
         dcrRowUpdates = (readOnly || dcrRowIndex != currRow)
                         ? null
                         : currRowUpdates;
      }

      /**
       * @see SpaceColumnRenderer#addData(com.goldencode.p2j.ui.BrowseColumnConfig, int, int, int)
       */
      @Override
      public void addData(BrowseColumnConfig col,
                          int colIndex,
                          int colWidth,
                          int scroll)
      {
         col.setColumn(screen().cursor().x - config.getColumn());

         if (!dcrReadOnly && 
              dcrFocused  && 
              dcrRowIndex == Math.max(currRow, 0) &&
              colIndex == currColumn)
         {
            // just cursor moving
            printSpaces(colWidth - Math.abs(scroll), colorCommon);
         }
         else
         {
            BaseDataType var =
               (dcrRowUpdates != null && dcrRowUpdates[colIndex] != null) ?
                  dcrRowUpdates[colIndex] :
                  dcrRow[colIndex];
            
            String display;
            try
            {
               if (!col.getDataType().equalsIgnoreCase(var.getTypeName()) &&
                   var.getTypeName().equals("CHARACTER"))
               {
                  // change column type and format
                  
                  DisplayFormat fmt = col.getDisplayFormat();
                  
                  int len = fmt.getScreenWidth();
                  boolean rightAligned = fmt.isRightPositioning() ||
                                         (fmt instanceof NumberFormat);
                               
                  if (len <= 0)
                     len = 1;
                  
                  col.setDataType("CHARACTER");
                  col.setFormat("x(" + len + ")");
                  
                  if (rightAligned)
                     col.forceRightPositioning();
               }
               
               String txt = col.getDisplayFormat().fromVar(var).toScreenValue();
               
               display = renderCell(txt, col, colWidth, scroll);
            }
            catch (DisplayFormatCheckException ex)
            {
               display = DisplayFormat.generateQuestions(colWidth - 
                                                         Math.abs(scroll));
            }

            Color clr = col.getDcolor();
            screen().append(display, getColor(clr));
         }
      }
   }
  
}
